use std::{collections::HashMap, vec};

use crate::parser::ASTPart;

const ASXVERSION: [u8; 3] = [0,1,0];

#[derive(Debug, Clone)]
pub struct Operation {
    pub opcode: u8,
    pub arg1: Option<u8>,
    pub arg2: Option<i64>,
    pub arg3: Option<u8>,
}

#[derive(Debug, Clone)]
struct RegisterState {
    id: u8,
    used: bool,
    variable: u32,
    last_used: usize,
}

#[derive(Debug, Clone)]
struct AllocateResult {
    register: u8,
    unbind_before: bool,
}

#[derive(Debug, Clone)]
struct Compiled {
    operations: Vec<Operation>,
    variables: Vec<Variable>,
    strings: HashMap<u32, String>,
    functions: HashMap<u32, Compiled>
}

#[derive(Debug, Clone)]
struct Variable {
    name: String,
    id: u32,
    start: usize,
    end: usize,
    no_end: bool
}

fn get_register_by_id(registers: &Vec<RegisterState>, id: u8) -> RegisterState {
    for register in registers {
        if register.id == id {
            return register.clone();
        }
    }
    return RegisterState {
        id: 0,
        used: false,
        variable: 0,
        last_used: 0,
    };
}

fn get_register_by_variable(registers: &Vec<RegisterState>, variable: Variable) -> RegisterState {
    for register in registers {
        if register.variable == variable.id {
            return register.clone();
        }
    }
    return RegisterState {
        id: 0,
        used: false,
        variable: 0,
        last_used: 0,
    };
}

fn set_register(registers: &mut Vec<RegisterState>, new_state: RegisterState) {
    if new_state.id == 0 {
        return;
    }
    for i in 0..registers.len() {
        if registers[i].id == new_state.id {
            registers[i] = new_state;
            return;
        }
    }
}

fn allocate_register(registers: &Vec<RegisterState>) -> AllocateResult {
    for register in registers {
        if register.id != 0 && !register.used {
            return AllocateResult {
                register: register.id,
                unbind_before: false,
            };
        }
    }
    //Find the one that wasn't used for the longest time
    let mut oldest: usize = 0;
    let mut oldest_register: u8 = 0;
    let mut oldest_temp: usize = 0;
    let mut oldest_temp_register: u8 = 0;
    for register in registers {
        if ((register.last_used < oldest) || (oldest == 0)) && register.id != 0 {
            oldest = register.last_used;
            oldest_register = register.id;
        }
        if ((register.last_used < oldest_temp) || (oldest_temp == 0)) && register.id != 0 && register.variable == 0 {
            oldest_temp = register.last_used;
            oldest_temp_register = register.id;
        }
    }
    if oldest_temp_register != 0 {
        return AllocateResult {
            register: oldest_temp_register,
            unbind_before: false,
        };
    }
    return AllocateResult {
        register: oldest_register,
        unbind_before: true,
    };
}

fn garbage_collect_registers(registers: &mut Vec<RegisterState>) {
    for register in registers {
        if register.used && register.variable == 0 {
            register.used = false;
            register.last_used = 0;
        }
    }
}

fn get_variable_by_name(variables: &Vec<Variable>, name: &str, position: usize) -> Option<Variable> {
    for variable in variables {
        if variable.name == name {
            if variable.start <= position && (variable.no_end || variable.end >= position) {
                return Some(variable.clone());
            }
        }
    }
    None
}

fn do_ast_op(ast_op: ASTPart, op_count: &mut usize, ops: &mut Vec<Operation>, variables: &mut Vec<Variable>, next_var_id: &mut u32, strings: &mut HashMap<u32, String>, next_string_id: &mut u32, functions: &mut HashMap<u32, Compiled>, next_function_id: &mut u32, registers: &mut Vec<RegisterState>) -> u8 {
    *op_count += 1;
    match ast_op {
        ASTPart::Number(num) => {
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 3, arg1: Some(reg.register), arg2: Some(num.value), arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::String(str) => {
            let str_id = *next_string_id;
            strings.insert(str_id, str.value);
            *next_string_id += 1;
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 1, arg1: Some(reg.register), arg2: Some(str_id as i64), arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::Boolean(bool) => {
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 4, arg1: Some(reg.register), arg2: Some(bool.value as i64), arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::Null(_) => {
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 6, arg1: Some(reg.register), arg2: None, arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::VarRead(var_read) => {
            if get_variable_by_name(variables, &var_read.variable, ops.len()).is_none() {
                panic!("Variable {} does not exist", var_read.variable);
            }
            let reg = get_register_by_variable(registers, get_variable_by_name(variables, &var_read.variable, ops.len()).expect("Variable should exist"));
            if reg.id == 0 {
                let reg = allocate_register(registers);
                if reg.unbind_before {
                    ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
                }
                ops.push(Operation { opcode: 2, arg1: Some(reg.register), arg2: Some(get_variable_by_name(variables, &var_read.variable, ops.len()).expect("Variable should exist").id as i64), arg3: None });
                set_register(registers, RegisterState { id: reg.register, used: true, variable: get_variable_by_name(variables, &var_read.variable, ops.len()).expect("Variable should exist").id, last_used: *op_count });
                return reg.register;
            } else {
                return reg.id;
            }
        },
        ASTPart::Function(func) => {
            let func_id = *next_function_id;
            functions.insert(func_id, compile_function(func.body, Some(func.args), registers, next_var_id));
            *next_function_id += 1;
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 5, arg1: Some(reg.register), arg2: Some(func_id as i64), arg3: None });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::Call(call) => {
            let func = do_ast_op(*call.function, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            for arg in call.args {
                let arg_reg = do_ast_op(arg, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
                ops.push(Operation { opcode: 28, arg1: Some(arg_reg), arg2: None, arg3: None });
                set_register(registers, RegisterState { id: arg_reg, used: false, variable: 0, last_used: 0 });
            }
            let ret_reg = allocate_register(registers);
            if ret_reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(ret_reg.register), arg2: None, arg3: None });
            }
            ops.push(Operation { opcode: 27, arg1: Some(func), arg2: Some(ret_reg.register as i64), arg3: None });
            return ret_reg.register;
        },
        ASTPart::If(if_part) => {
            let condition_reg = do_ast_op(*if_part.condition, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            ops.push(Operation { opcode: 24, arg1: Some(condition_reg), arg2: Some(condition_reg as i64), arg3: None });
            
            //Update the lastif variable
            if get_variable_by_name(variables, "__LASTIF", ops.len()).is_none() {
                variables.push(Variable { name: String::from("__LASTIF"), id: *next_var_id, start: ops.len(), end: 0, no_end: true });
                *next_var_id += 1;
            }
            ops.push(Operation { opcode: 7, arg1: Some(condition_reg), arg2: Some(get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id as i64), arg3: None });
            set_register(registers, RegisterState { id: condition_reg, used: true, variable: get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id, last_used: *op_count });

            let op_placeholder = ops.len();
            ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
            let mut fake_vars: Vec<Variable> = vec![];
            for if_op in if_part.body {
                do_ast_op(if_op, op_count, ops, &mut fake_vars, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            }
            for fake_var in fake_vars {
                variables.push(Variable { name: fake_var.name, id: fake_var.id, start: fake_var.start, end: ops.len()-1, no_end: false });
            }
            ops[op_placeholder] = Operation {
                opcode: 26,
                arg1: Some(condition_reg),
                arg2: Some(ops.len() as i64),
                arg3: None,
            };

            //Unbind the lastif variable
            if get_register_by_id(registers, condition_reg).variable == get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id {
                ops.push(Operation { opcode: 8, arg1: Some(condition_reg), arg2: None, arg3: None });
                set_register(registers, RegisterState { id: condition_reg, used: false, variable: 0, last_used: 0 });
            }
            garbage_collect_registers(registers);
        },
        ASTPart::Else(else_part) => {
            if get_variable_by_name(variables, "__LASTIF", ops.len()).is_none() {
                panic!("Else used without an if statement before it");
            }
            let reg = get_register_by_variable(registers, get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").clone());
            let else_condition_reg;
            if reg.id == 0 {
                let reg = allocate_register(registers);
                if reg.unbind_before {
                    ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
                }
                ops.push(Operation { opcode: 2, arg1: Some(reg.register), arg2: Some(get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id as i64), arg3: None });
                set_register(registers, RegisterState { id: reg.register, used: true, variable: get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id, last_used: *op_count });
                else_condition_reg = reg.register;
            } else {
                else_condition_reg = reg.id;
            }
            ops.push(Operation { opcode: 24, arg1: Some(else_condition_reg), arg2: Some(else_condition_reg as i64), arg3: None });

            let op_placeholder = ops.len();
            ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
            let mut fake_vars: Vec<Variable> = vec![];
            for else_op in else_part.body {
                do_ast_op(else_op, op_count, ops, &mut fake_vars, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            }
            for fake_var in fake_vars {
                variables.push(Variable { name: fake_var.name, id: fake_var.id, start: fake_var.start, end: ops.len()-1, no_end: false });
            }
            ops[op_placeholder] = Operation {
                opcode: 26,
                arg1: Some(else_condition_reg),
                arg2: Some(ops.len() as i64),
                arg3: None,
            };
            //Unbind the lastif variable
            if get_register_by_id(registers, else_condition_reg).variable == get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id {
                ops.push(Operation { opcode: 8, arg1: Some(else_condition_reg), arg2: None, arg3: None });
                set_register(registers, RegisterState { id: else_condition_reg, used: false, variable: 0, last_used: 0 });
            }
            garbage_collect_registers(registers);
        },
        ASTPart::ElseIf(elseif_part) => {
            if get_variable_by_name(variables, "__LASTIF", ops.len()).is_none() {
                panic!("Else if used without an if statement before it");
            }
            let reg = get_register_by_variable(registers, get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").clone());
            let else_condition_reg;
            if reg.id == 0 {
                let reg = allocate_register(registers);
                if reg.unbind_before {
                    ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
                }
                ops.push(Operation { opcode: 2, arg1: Some(reg.register), arg2: Some(get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id as i64), arg3: None });
                set_register(registers, RegisterState { id: reg.register, used: true, variable: get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id, last_used: *op_count });
                else_condition_reg = reg.register;
            } else {
                else_condition_reg = reg.id;
            }
            ops.push(Operation { opcode: 24, arg1: Some(else_condition_reg), arg2: Some(else_condition_reg as i64), arg3: None });

            let condition_reg = do_ast_op(*elseif_part.condition, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            ops.push(Operation { opcode: 16, arg1: Some(else_condition_reg), arg2: Some(condition_reg as i64), arg3: Some(condition_reg) });
            ops.push(Operation { opcode: 24, arg1: Some(condition_reg), arg2: Some(condition_reg as i64), arg3: None });

            //Update the lastif variable
            ops.push(Operation { opcode: 7, arg1: Some(condition_reg), arg2: Some(get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id as i64), arg3: None });
            set_register(registers, RegisterState { id: condition_reg, used: true, variable: get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id, last_used: *op_count });

            let op_placeholder = ops.len();
            ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
            let mut fake_vars: Vec<Variable> = vec![];
            for elseif_op in elseif_part.body {
                do_ast_op(elseif_op, op_count, ops, &mut fake_vars, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            }
            for fake_var in fake_vars {
                variables.push(Variable { name: fake_var.name, id: fake_var.id, start: fake_var.start, end: ops.len()-1, no_end: false });
            }
            ops[op_placeholder] = Operation {
                opcode: 26,
                arg1: Some(condition_reg),
                arg2: Some(ops.len() as i64),
                arg3: None,
            };
            //Unbind the lastif variable
            if get_register_by_id(registers, condition_reg).variable == get_variable_by_name(variables, "__LASTIF", ops.len()).expect("__LASTIF should exist").id {
                ops.push(Operation { opcode: 8, arg1: Some(condition_reg), arg2: None, arg3: None });
                set_register(registers, RegisterState { id: condition_reg, used: false, variable: 0, last_used: 0 });
            }
            garbage_collect_registers(registers);
        },
        ASTPart::While(while_part) => {
            let start = ops.len();
            let condition_reg = do_ast_op(*while_part.condition, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            ops.push(Operation { opcode: 24, arg1: Some(condition_reg), arg2: Some(condition_reg as i64), arg3: None });
            
            let op_placeholder = ops.len();
            let mut breaks: Vec<usize> = vec![];
            let mut continues: Vec<usize> = vec![];
            ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
            for while_op in while_part.body {
                match while_op {
                    ASTPart::Break(_) => {
                        breaks.push(ops.len());
                        ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
                    },
                    ASTPart::Continue(_) => {
                        continues.push(ops.len());
                        ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
                    },
                    _ => {
                        do_ast_op(while_op, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
                    }
                }
            }
            ops.push(Operation { opcode: 25, arg1: None, arg2: Some(start as i64), arg3: None });
            ops[op_placeholder] = Operation {
                opcode: 26,
                arg1: Some(condition_reg),
                arg2: Some(ops.len() as i64),
                arg3: None,
            };
            for brk in breaks {
                ops[brk] = Operation {
                    opcode: 25,
                    arg1: None,
                    arg2: Some(ops.len() as i64),
                    arg3: None,
                };
            }
            for cont in continues {
                ops[cont] = Operation {
                    opcode: 25,
                    arg1: None,
                    arg2: Some(start as i64),
                    arg3: None,
                };
            }
        },
        ASTPart::Break(_) => {
            panic!("Unexpected break outside of loop at {}", op_count);
        },
        ASTPart::Continue(_) => {
            panic!("Unexpected continue outside of loop at {}", op_count);
        },
        ASTPart::For(for_part) => {
            do_ast_op(*for_part.init, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            let start = ops.len();
            let condition_reg = do_ast_op(*for_part.condition, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            ops.push(Operation { opcode: 24, arg1: Some(condition_reg), arg2: Some(condition_reg as i64), arg3: None });

            let op_placeholder = ops.len();
            let mut breaks: Vec<usize> = vec![];
            let mut continues: Vec<usize> = vec![];
            ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
            for for_op in for_part.body {
                match for_op {
                    ASTPart::Break(_) => {
                        breaks.push(ops.len());
                        ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
                    },
                    ASTPart::Continue(_) => {
                        continues.push(ops.len());
                        ops.push(Operation { opcode: 0, arg1: None, arg2: None, arg3: None });
                    },
                    _ => {
                        do_ast_op(for_op, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
                    }
                }
            }
            do_ast_op(*for_part.update, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            
            ops.push(Operation { opcode: 25, arg1: None, arg2: Some(start as i64), arg3: None });
            ops[op_placeholder] = Operation {
                opcode: 26,
                arg1: Some(condition_reg),
                arg2: Some(ops.len() as i64),
                arg3: None,
            };
            for brk in breaks {
                ops[brk] = Operation {
                    opcode: 25,
                    arg1: None,
                    arg2: Some(ops.len() as i64),
                    arg3: None,
                };
            }
            for cont in continues {
                ops[cont] = Operation {
                    opcode: 25,
                    arg1: None,
                    arg2: Some(start as i64),
                    arg3: None,
                };
            }
        },
        ASTPart::Return(ret) => {
            let ret_reg = do_ast_op(*ret.value, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            ops.push(Operation { opcode: 29, arg1: Some(ret_reg), arg2: None, arg3: None });
        },
        ASTPart::Operation(op) => {
            let left_reg = do_ast_op(*op.left, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            let right_reg = do_ast_op(*op.right, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            let reg = allocate_register(registers);
            if reg.unbind_before {
                ops.push(Operation { opcode: 8, arg1: Some(reg.register), arg2: None, arg3: None });
            }
            let opcode = match op.operator.as_str() {
                "+" => 10,
                "-" => 11,
                "*" => 12,
                "/" => 13,
                "^" => 14,
                "%" => 15,
                "&" => 16,
                "|" => 17,
                "==" => 18,
                "!=" => 19,
                ">" => 20,
                ">=" => 21,
                "<" => 22,
                "<=" => 23,
                "!" => {
                    ops.push(Operation { opcode: 24, arg1: Some(left_reg), arg2: Some(reg.register as i64), arg3: None });
                    set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
                    return reg.register;
                },
                _ => panic!("Unknown operator {}", op.operator),
            };
            ops.push(Operation { opcode, arg1: Some(left_reg), arg2: Some(right_reg as i64), arg3: Some(reg.register) });
            set_register(registers, RegisterState { id: reg.register, used: true, variable: 0, last_used: *op_count });
            return reg.register;
        },
        ASTPart::VarUpdate(upd) => {
            if get_variable_by_name(variables, &upd.variable, ops.len()).is_none() {
                panic!("Variable {} does not exist", upd.variable);
            }
            let reg = get_register_by_variable(registers, get_variable_by_name(variables, &upd.variable, ops.len()).expect("Variable should exist").clone());
            if reg.id != 0 {
                ops.push(Operation { opcode: 8, arg1: Some(reg.id), arg2: None, arg3: None });
                set_register(registers, RegisterState { id: reg.id, used: false, variable: 0, last_used: 0 });
            }
            let value_reg = do_ast_op(*upd.value, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            ops.push(Operation { opcode: 7, arg1: Some(value_reg), arg2: Some(get_variable_by_name(variables, &upd.variable, ops.len()).expect("Variable should exist").id as i64), arg3: None });
            set_register(registers, RegisterState { id: value_reg, used: true, variable: get_variable_by_name(variables, &upd.variable, ops.len()).expect("Variable should exist").id, last_used: *op_count });
            garbage_collect_registers(registers);
        },
        ASTPart::Assigment(asign) => {
            if get_variable_by_name(variables, &asign.variable, ops.len()).is_some() {
                panic!("Variable {} already exists", asign.variable);
            }
            let reg = do_ast_op(*asign.value, op_count, ops, variables, next_var_id, strings, next_string_id, functions, next_function_id, registers);
            variables.push(Variable { name: asign.variable.clone(), id: *next_var_id, start: ops.len(), end: 0, no_end: true });
            *next_var_id += 1;
            ops.push(Operation { opcode: 7, arg1: Some(reg), arg2: Some(get_variable_by_name(variables, &asign.variable, ops.len()).expect("Variable should exist").id as i64), arg3: None });
            set_register(registers, RegisterState { id: reg, used: true, variable: get_variable_by_name(variables, &asign.variable, ops.len()).expect("Variable should exist").id, last_used: *op_count });
            garbage_collect_registers(registers);
        },
        _ => {}
    }
    return 0;
}

fn compile_function(ast: Vec<ASTPart>, args: Option<Vec<String>>, registers: &mut Vec<RegisterState>, next_var_id: &mut u32) -> Compiled {
    let mut ops: Vec<Operation> = vec![];

    let mut variables: Vec<Variable> = vec![];
    let mut strings: HashMap<u32, String> = HashMap::new();
    let mut next_string_id: u32 = 1;
    let mut functions: HashMap<u32, Compiled> = HashMap::new();
    let mut next_function_id: u32 = 1;

    match args {
        Some(arg_list) => {
            for arg in arg_list {
                if get_variable_by_name(&variables, &arg, 0).is_some() {
                    panic!("Argument {} already exists", arg);
                }
                variables.push( Variable { name: arg, id: *next_var_id, start: 0, end: 0, no_end: true });
                *next_var_id += 1;
            }
        },
        None => {}
    }

    let mut op_count = 0;
    for ast_op in ast {
        do_ast_op(ast_op, &mut op_count, &mut ops, &mut variables, next_var_id, &mut strings, &mut next_string_id, &mut functions, &mut next_function_id, registers);
    }
    ops.push(Operation { opcode: 29, arg1: Some(0), arg2: None, arg3: None });
    for var in &mut variables {
        if var.no_end {
            var.end = ops.len() - 1;
        }
    }

    return Compiled {
        operations: ops,
        variables,
        strings,
        functions,
    };
}

fn append_be_num(output: &mut Vec<u8>, size: usize, num: usize) { //appends a big-endian u number to the output
    for i in (0..size).rev() {
        output.push(((num >> (i * 8)) & 0xFF) as u8);
    }
}

fn append_string(output: &mut Vec<u8>, input: String, add_len: bool) {
    let bytes = input.as_bytes();
    if add_len {
        append_be_num(output, 4, bytes.len());
    }
    output.extend_from_slice(bytes);
}

fn append_binary(output: &mut Vec<u8>, input: String) {
    let mut inp = input.clone();
    while inp.len() >= 8 {
        let byte = bin_to_num(inp[..8].to_string());
        output.push(byte as u8);
        inp = inp[8..].to_string();
    }
}

fn num_to_bin(number: usize, size: usize) -> String {
    let mut bin = String::new();
    for i in (0..size).rev() {
        if (number >> i) & 1 == 1 {
            bin.push('1');
        } else {
            bin.push('0');
        }
    }
    return bin;
}

fn bin_to_num(bin: String) -> usize {
    let mut num = 0;
    for (i, c) in bin.chars().rev().enumerate() {
        if c == '1' {
            num += 1 << i;
        }
    }
    return num;
}

fn compile_body(compiled: Compiled, fpos: &mut usize) -> Vec<u8> {
    let mut output: Vec<u8> = vec![];
    let mut additional: Vec<Vec<u8>> = vec![];
    //function
    //function headers
    append_be_num(&mut output, 4, compiled.variables.len());
    for var in compiled.variables {
        append_string(&mut output, var.name, true);
        append_be_num(&mut output, 3, var.id as usize);
        append_be_num(&mut output, 4, var.start);
        append_be_num(&mut output, 4, var.end);
    }
    append_be_num(&mut output, 4, compiled.strings.len());
    for strs in compiled.strings.keys() {
        append_be_num(&mut output, 3, *strs as usize);
        append_string(&mut output, compiled.strings[strs].clone(), true);
    }
    append_be_num(&mut output, 4, compiled.functions.len());
    for funcs in compiled.functions.keys() {
        *fpos += 1;
        append_be_num(&mut output, 3, *funcs as usize);
        append_be_num(&mut output, 4, *fpos);
        additional.push(compile_body(compiled.functions[funcs].clone(), fpos));
    }
    //function body
    append_be_num(&mut output, 4, compiled.operations.len());
    for ops in compiled.operations {
        let mut op_bin = num_to_bin(ops.opcode.into(), 5);
        if let Some(arg1) = ops.arg1 {
            op_bin.push_str(&num_to_bin(arg1.into(), 4));
        } else {
            op_bin.push_str(&num_to_bin(0, 4));
        }
        if let Some(arg2) = ops.arg2 {
            op_bin.push_str(&num_to_bin(arg2 as usize, 64));
        } else {
            op_bin.push_str(&num_to_bin(0, 64));
        }
        if let Some(arg3) = ops.arg3 {
            op_bin.push_str(&num_to_bin(arg3.into(), 4));
        } else {
            op_bin.push_str(&num_to_bin(0, 4));
        }
        op_bin.push_str(&num_to_bin(0, 3));
        append_binary(&mut output, op_bin);
    }
    for addition in additional {
        output.extend_from_slice(&addition);
    }

    return output;
}

pub fn compile(ast: Vec<ASTPart>) -> Vec<u8> {
    let mut next_var_id: u32 = 1;
    let mut registers: Vec<RegisterState> = vec![];
    for i in 0..17 {
        registers.push(RegisterState {
            id: i as u8,
            used: false,
            variable: 0,
            last_used: 0,
        });
    }

    let compiled = compile_function(ast, None, &mut registers, &mut next_var_id);

    let mut output: Vec<u8> = vec![];
    //doctype specifier
    append_string(&mut output, String::from("ASX"), false);
    //version
    output.push(ASXVERSION[0]);
    output.push(ASXVERSION[1]);
    output.push(ASXVERSION[2]);
    //functions
    let mut fpos = 0;
    let body = compile_body(compiled, &mut fpos);
    output.extend_from_slice(&body);

    return output;
}